/* PIC file loader for MINIX-X68
 *
 * ver 0.01 by Akira-Y.
 * modified by murao.
 * bug-fix  by kat.
 */

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

#define dprintf(p) /*printf p /* */

#ifndef L_SET
#define L_SET  (0)
#endif

#define SIZE_OF_X		512
#define SIZE_OF_Y		512

#define SIZE_OF_BUFF		32000
#define VRAM_SIZE	  	262144L

unsigned short *vram_buff;

void (*pset_ent)(); 	/* put a pixel */
int  (*point_ent)();	/* get a pixel */
int  vram_buffered; 	/* option flag -b specified */
int  vram;		/* fd for /dev/vram */
int  vram_opened;
int  handle;		/* fd for specified .pic file */
int  bit_len;		/* *buff_p's effective bit length */
char *buff_p;		/* buffer's read pointer */
int  buff_len;		/* buff rest */
int  x_wid, y_wid;	/* pic's screen width */

#if 0
char *buff;		/* file read buff */
#else
char buff[SIZE_OF_BUFF];
#endif

struct {
	int color;
	int next;
	int prev;
} table[128];		/* color cache table */
int color_p;		/* index of color table */

void	ginit(), expand(), header_read(), color_cache_init();
void	pset_buf(), pset_nobuf();
void	expand_chain(), buff_next(), error(), usage();
int	read_color(), point_buf(), point_nobuf();
long	read_len(), bit_load();

void
main( argc, argv )
int argc;
char *argv[];
{
	char *file;
	vram_opened = vram_buffered = 0;
	pset_ent = pset_nobuf;
	point_ent = point_nobuf;

	printf("PIC loader ver 0.01 by Akira-Y. modified by murao.\n");
	if( argc == 1 ) usage();
	else
	{
		int i;

		for( i = 1 ; i < argc ; i++ )
		{
			if( *argv[i] == '-' )
			{
				switch(*(argv[i]+1))
				{
				case 'b':
					vram_buffered = 1;
					pset_ent = pset_buf;
					point_ent = point_buf;
					break;
					
				case 'h':
					usage();
					break;
					
				default:
					error("unknown option. try -h for more information.");
				}
			}
			else file = argv[i];
		}
	}

	if( ( handle = open(file, O_RDONLY) ) < 0 )
		error("cannot open the file");
	if( ( vram = open("/dev/vram", O_RDWR) ) < 0 )
		error("cannot open /dev/vram");
	else
		vram_opened = 1;
	bit_len = buff_len = 0;
	ginit();
	color_cache_init();
	header_read();
	expand();
	close(handle);
	if( vram_buffered )
	{
		char *vram_buff_p = (char*)vram_buff;
		long size = VRAM_SIZE * 2L;

		lseek( vram, 0, L_SET );
		while( size > 32767 )
		{
			write( vram, vram_buff_p, 32766 );
			size -= 32766L;
			vram_buff_p += 32766;
		}
		if( size > 0 ) write( vram, (char *)vram_buff, size );
	}
	if( vram_opened ) close(vram);
}

void
header_read()
{
	long bit_load();
	int c;

	dprintf(("reading header...\n"));
	if((int)bit_load(8) != 'P') error("no pic file");
	if((int)bit_load(8) != 'I') error("no pic file");
	if((int)bit_load(8) != 'C') error("no pic file");
	while((c = (int)bit_load(8)) != 26 ) putchar(c);
	while((int)bit_load(8) != 0 );

	/* pic mode check */
	if((int)bit_load(16) != 0) error("no pic file");

	/* pic color bit length check */
	if((int)bit_load(16) != 15) error("no pic file");
	if((x_wid = (int)bit_load(16)) > SIZE_OF_X) error("size over");
	if((y_wid = (int)bit_load(16)) > SIZE_OF_Y)
	{
		y_wid = SIZE_OF_Y;
	}
}

void
expand()
{
	int x, y, c, a;
	long l;

	dprintf(("expanding...\n"));
	x = -1;
	y = c = 0;
	for(;;)
	{
		l = read_len();
		while(--l)
		{
			if(++x == x_wid)
			{
				if(++y == y_wid) return;
				x = 0;
			}
			if((a = (*point_ent)(x,y)) != 0 ) c = a & 0xfffe;
			(*pset_ent)(x, y, c);
		}
		if(++x == x_wid)
		{
			if(++y == y_wid) return;
			x = 0;
		}
		c = read_color();
		(*pset_ent)(x, y, c);
		if( bit_load(1) != 0 ) expand_chain(x, y, c | 1);
	}
}

void
expand_chain(x, y, c)
	int x, y, c;
{
	int y_over = 1;

	for(;;)
	{
		switch(bit_load(2))
		{
		  case 0:
			if(bit_load(1) == 0) return;
			if(bit_load(1) == 0) x -= 2;
			else x += 2;
			break;
		  case 1: x--; break;
		  case 2:      break;
		  case 3: x++; break;
		}
		if( x >= x_wid ) return;
		if( ++y >= y_wid ) y_over = 0;
		if( y_over ) (*pset_ent)(x, y, c);
	}
}

long
read_len()
{
	int a = 1;

	while(bit_load(1) != 0) a++;
	return(bit_load(a) + (1<<a) - 1);
}

int
read_color()
{
	long bit_load();

	if( bit_load(1) == 0 ) return(new_color((int)bit_load(15)));
	else return(get_color((int)bit_load(7)));
}

void
color_cache_init()
{
	int i;

	dprintf(("initializing color cache...\n"));
	for( i = 0 ; i < 128 ; i++ )
	{
		table[i].color = 0;
		table[i].prev = i + 1;
		table[i].next = i - 1;
	}
	table[127].prev = 0;
	table[0].next = 127;
	color_p = 0;
}

int
new_color(c)
int c;
{
	color_p = table[color_p].prev;
	table[color_p].color = c;
	return(2*c);
}

int
get_color(idx)
int idx;
{
	if(color_p != idx)
	{
		table[table[idx].prev].next = table[idx].next;
		table[table[idx].next].prev = table[idx].prev;

		table[table[color_p].prev].next = idx;
		table[idx].prev = table[color_p].prev;
		table[color_p].prev = idx;
		table[idx].next = color_p;
		color_p = idx;
	}
	return(table[idx].color * 2);
}

#ifndef ASM_BITLOAD
long
bit_load(size)
	int size;
{
	int  i;
	long a = 0;	/* return value */
	
	dprintf(("bit_load(%d)\n", size));
	while (size > bit_len)
	{
		for (i = 0; i < bit_len; i++)
		{
			a += a;					/* shift left */
			if(*buff_p < 0) a++;	/* a |= 1 */
			*buff_p += *buff_p; 	/* shift left */
			size--;
		}
		buff_next();
	}
	
	for (i = 0; i < size; i++)
	{
		a += a;					/* shift left */
		if(*buff_p < 0) a++;  	/* a |= 1 */
		*buff_p += *buff_p; 	/* shift left */
		bit_len--;
	}
	return(a);
}
#endif

void
buff_next()
{
	dprintf(("buff_next()\n"));
	if( buff_len == 0 )
	{
		if((buff_len = read(handle, buff, SIZE_OF_BUFF)) <= 0 )
			error("file read error");
		buff_p = buff;
	}
	else buff_p++;
	buff_len--;
	bit_len = 8;
}

void
error(s)
char *s;
{
	puts(s);
	if(vram_opened) close(vram);
	exit(1);
}

void
usage()
{
	fprintf( stderr, "usage : picl [<options>] <file>\n" );
	fprintf( stderr, "option -b vram buffered (needs over 512KB memory)\n" );
	fprintf( stderr, "       -h show this message.\n" );
	exit(0);
}

/***** system dependent subroutines *****/

int
point_buf(x, y)
	int x, y;
{
	off_t pos = (long)y * 512L + (long)x;
	return (int)vram_buff[pos];
}

int
point_nobuf(x, y)
	int x, y;
{
	off_t pos;
	unsigned short col;
	
	pos = (long)y * 512L + (long)x;
	pos *= 2L;
	lseek( vram, pos, L_SET );
	if( read( vram, &col, sizeof(unsigned short)) <= 0 )
		error("failed to read out /dev/vram");
	return (int)col;
}

void
pset_buf(x, y, c) /* called when (vram_buffered) */
	int x, y, c;
{
	long pos = (long)y * 512L + (long)x;
	vram_buff[pos] = (unsigned short)c;
}

void
pset_nobuf(x, y, c) /* called when (!vram_buffered) */
	int x, y, c;
{
	unsigned short col;
	static off_t pos;
	static int oldx = 9999, oldy = 9999;
	
	++oldx;
	if (oldx == x && oldy == y)
	{
		pos += 2;
	}	
	else
	{
		oldx = x;
		oldy = y;
		pos = (long)y * 512L + (long)x;
		pos *= 2L;
	}
	lseek( vram, pos, L_SET );
	col = (unsigned short)c;
	if( write( vram, &col, sizeof(unsigned short)) <= 0 )
		error("error on pset");
}

unsigned short vramBuffer[SIZE_OF_Y][SIZE_OF_X];
void
ginit(x, y)
	int x, y;
{
	register long i;
	
	dprintf(("inisializing buffer..\n"));
	vram_buff = (unsigned short*)vramBuffer;
#if 1			
	for (i = 0; i < SIZE_OF_Y; i++)
		memset(vramBuffer[i], 0, SIZE_OF_X * sizeof(unsigned short));
#else
	for(i = 0 ; i < VRAM_SIZE ; i++)
		vram_buff[i] = 0;
#endif
}
